"use strict";
this.name	= "ils";
this.author	= "Norby";
this.copyright	= "2016 Norby";
this.description= "Instrument Landing System";
this.licence	= "CC BY-NC-SA 4.0";

//internal properties, should not touch
this.$F = null; //FCB pointer
this.$L = null; //last ILS user ship
this.$N = null; //wide NPC in last stage
this.$P = null; //previous heading in FCB
this.$S = null; //the target station
this.$T = null; //Towbar worldScript pointer
this.$Targets = null; //Result of scanner check
this.$Timer = null; //Timer for scanner check
this.$V = null; //visualEffect for the white flashing marker


//worldscript events
this.startUp = function() {
	this.$T = worldScripts.towbar;
	if( !isValidFrameCallback( this.$F ) )
		this.$F = addFrameCallback( this.$FCB.bind(this) );
        this.$Timer = new Timer(this, this.$Timed.bind(this), 1, 1);//scanner check in every seconds
}

this.equipmentDamaged = this.equipmentDestroyed = function(equipment) {
	if(equipment == "EQ_ILS") this.$ClearILS(); //stop ILS if working
}

/* another oxp cause a bug which turn off ILS immediately in the next lines so do not use this event
this.shipTargetLost = function() {
	log(this.name, "ILS OFF 1");//debug
	this.$ClearILS();
}
*/

this.shipWillDockWithStation = this.shipWillEnterWitchspace = this.shipDied = function() {
	this.$ClearILS(true);//nomsg
}

this.shipWillLaunchFromStation = function(station) {
        if( this.$L && this.$L != player.ship && this.$L.isValid && this.$L.isInSpace ) {
                this.$L.desiredSpeed = 0; //stop landing NPC when player launch
        }
        this.$ClearILS(true);
}

//ILS methods
this.$ClearILS = function(nomsg) {
	this.$N = null;
	this.$P = null;
	this.$S = null;
	if( this.$V && this.$V.isVisualEffect ) {
		this.$V.remove();
		if( !nomsg && this.$L == player.ship )
			player.consoleMessage("ILS off");
	}
	this.$V = null;
	this.$L = null;
//	if( isValidFrameCallback( this.$F ) ) removeFrameCallback( this.$F );
//	log(this.name, "ILS OFF");//debug
}

this.$NPC = function( t ) { //target is a landing NPC ship?
	if( t && t.isPiloted && t.maxSpeed > 0 && t != player.ship
		&& ( t.AI == "dockingAI.plist" || t.dockingInstructions )
		&& t.target && t.target.isStation && t.target.maxSpeed == 0 ) {
                return(true);
//                 //wide NPC need rolling target station to turn on ILS - disabled since can roll
//                 if( t.target.roll != 0 ) return(true);
//                 var u = t.target.subEntities;
//                 var i = 0;
//                 if( u ) while( u[i] && !u[i].isDock ) i++;
//                 if( u && u[i] && u[i].isDock &&
//                         t.boundingBox.x > Math.min( u[i].boundingBox.x, u[i].boundingBox.y ) ) {
// 		        return(true); //not wide so fit into without roll
//                 }
        }
	return(false);
}

this.$FCB = function( delta ) {
//	var w = worldScripts.ils;
//	player.consoleMessage(this.$S+" "+delta); //debug
	var p = player.ship;
	if( !p || !p.isValid ) return;
	var pp = p.position;
	var t = p.target;
	var tis = false; //target is station but not a carrier?
	var npc = false; //there is a landing NPC ship?
	var near = 10000; //NPCs use ILS within this distance
	var noinjdist = 2000; //switch off injectors within this distance
	var maxw = 64; //safe max. width of ship to fit into the dock height without roll

	if( t && ( t.isStation && t.dataKey.indexOf("carrier") == -1 || t == this.$L )
		&& !p.missilesOnline //not in missile targeting mode
		&& ( p.alertCondition != 3 || !p.weaponsOnline ) //not in red alert with online weapons
		&& p.equipmentStatus("EQ_ILS") == "EQUIPMENT_OK" ) { //has ILS
		if( !this.$L || this.$L == p //ILS is not used by NPCs at the moment
			|| !this.$L.isValid || !this.$L.isInSpace ) { //or docked/destroyed
			if( t != this.$L ) tis = true;
			else if( this.$L.target && this.$L.target.isStation ) {
				 tis = true; //set target to the station when npc is docked
				 if( p.target != this.$L.target )
				 	t = p.target = this.$L.target; 
			}
		} else {
			if( p.target != this.$L ) {
				if( pp.distanceTo( this.$L.position ) < p.scannerRange ) {
					//prevent repeated messages when target is out of range
					p.target = this.$L;
				}
				player.consoleMessage("ILS locked by "+this.$L.displayName, 5);
			}
		}
	}

	if( player.dockingClearanceStatus == "DOCKING_CLEARANCE_STATUS_GRANTED"
                || player.dockingClearanceStatus == "DOCKING_CLEARANCE_STATUS_TIMING_OUT" ) {
                    if( this.$L && this.$L != p && this.$L.isValid ) {
                        this.$L.desiredSpeed = 0; //stop NPC when player has docking clearance
                    }
        } else if( this.$L && this.$L != p && this.$L.isValid && this.$L.isInSpace
		&& this.$NPC( this.$L ) ) { //still target to the station? (not if attacked)
		npc = true; //select last npc until is not docked or destroyed
		t = this.$L;
	} else if( !tis ) { //an NPC can use ILS when is not used at the moment
		t = null;
                var targets = this.$Targets; //filled by checkScanner() in timer
		if( targets && targets.length > 0 ) {
			var bestsd = 0; //best speed/distance mean fastest landing so far
			for( var i = 0; i < targets.length; i++ ) {
				var n = targets[i];
				if( this.$NPC( n ) ) {
					//ILS on the ship which probably arrive first
					var nd = n.position.distanceTo(n.target.position);
					var sd = n.maxSpeed/nd;
					if( nd < near && nd > noinjdist && n.fuel > 0.2
						&& n.equipmentStatus("EQ_FUEL_INJECTION")
							== "EQUIPMENT_OK" ) {
						if( n.injectorSpeedFactor > 0 )
							sd = n.injectorSpeedFactor * sd;
						else sd = 7 * sd; //for Oolite 1.80
					}
					if( nd < near && ( !t || sd > bestsd ) //first or faster
						&& n.boundingBox.x < 210 ) {//exclude bigships
						npc = true;
						t = n; //only one NPC can use ILS at the same time
						this.$N = null;
						t.awardEquipment("EQ_ILS");
						bestsd = sd;
					} else if( nd > near ) { //no ILS, just a speedup
						if( n.injectorSpeedFactor > 0 )
							n.desiredSpeed = n.maxSpeed
								* n.injectorSpeedFactor;
						else n.desiredSpeed = 7 * n.maxSpeed;	
					}
				}
			}
//			if(t) player.consoleMessage(t.name+" "+t.entityPersonality+" ");//debug
		}
	}
	
	if( t && t.isValid && ( tis || npc ) ) {
		if( npc ) { //NPC ILS
			p = t;
			t = t.target;
		}

		var warning = false;
		if( this.$S != t || this.$L != p ) { //first lock on this target
			this.$S = t; //target station
			var m = "";
			if( p != player.ship ) {
				m = " of "+p.name;
//				if( !player.ship.target ) player.ship.target = p; - confusing, removed
			} else warning = true;
			if( !player.ship.docked )
				player.consoleMessage("ILS"+m+" set to "+t.displayName, 5);
//			log(this.name, "ILS set to: "+t);//debug
		}
		this.$L = p; //last ILS user ship
		
		var ph = p.heading;
		pp = p.position; //must set again for NPCs
		if( !this.$P ) this.$P = ph;
		var s = this.$S; //station
		var mp = p.maxPitch;
		var mr = p.maxRoll;
		var my = p.maxYaw;
//		log(this.name, "Station: "+s);//debug

		var u = s.subEntities;
		var i = 0;
		if( u ) while( u[i] && !u[i].isDock ) i++;
		var dp = s.position; //determine dock position
		var dr = s.vectorRight; //save for wide npc
		var dz = s.boundingBox.z;
		if( u && u[i] && u[i].isDock ) {
//			var dd = u[i].position.magnitude(); //dock dist. from the center of the station
//			dp = s.position.add( s.heading.direction().multiply(dd) ); //old formula
			dp = s.position.add( //better formula for non-centered docks
				Vector3D( s.heading.direction().multiply( u[i].position.z ),
				s.vectorUp.direction().multiply( u[i].position.y ),
				dr.direction().multiply( u[i].position.x ) ) );
			var x = u[i].boundingBox.x, y = u[i].boundingBox.y;
			if( y > x ) { //rotated dock (coriolis, ico)
				maxw = x;
				dr = s.vectorUp;
			} else maxw = y; //dodo, etc.
			dz = u[i].boundingBox.z; //save for wide npc
		}
                var dd = pp.distanceTo( dp ); //dock distance, used far below for NPCs also
                if( warning ) {
                        if( p.boundingBox.x < maxw )
                                player.consoleMessage("You can dock without roll", 5);
                        else if( dd < noinjdist )
                            player.consoleMessage("Auto roll is active to fit into the dock", 5);
                        else player.consoleMessage("Auto roll will be active before the dock", 5);
                }
		var d = dd / 2;
		var w = s.position.subtract( pp ).direction();
		var an = s.heading.angleTo(w);
//		player.consoleMessage(d +" "+ s.heading.angleTo(w)); //debug

		if( d < 100 || ( an < 1.57 && d < 300 ) ) //fix for rock hermit
			d = -500; //for sure, do not turn back if crossed the point in the dock
		else {
			if( an < 3.14 ) { //player is not before the station yet
				var na = 3.14-an;
				na = na * na;
				d += na * 2000;
				//extend distance if apporaching behind from the station
			}
			d = Math.min( 9500, d ); //max. at the Nav Buoy
		}
		var ap = dp.add( s.heading.direction().multiply( d ) ); //aimed position

		var v = ap.subtract( pp ).direction();
		if( this.$V && this.$V.isVisualEffect ) this.$V.position = ap;
		else this.$V = system.addVisualEffect("ils-marker", ap);
		
		var angle = ph.angleTo(v);

		var ato = 0;
		if(ph && this.$P) ato = ph.angleTo(this.$P);//angle to the Previous Heading
		if( ato < 0.005 && angle > 0.0015
			 //steer if no manual steering and not in 0.1 degree
			&& p.speed > 1 ) { //and player is not stopped
			//if the above ato value lower then can not start steering in my intel atom netbook
//			player.consoleMessage(this.$P +" vs "+ ph); //debug
			var r = Math.min(p.maxPitch*delta, angle/12);//half max turn/step and not too accutate

			if( !npc && this.$T && this.$T.$TowbarShip
				&& this.$T.$TowbarShip.isValid ) {
				var ma = Math.min( 2, p.mass / this.$T.$TowbarShip.mass ); ///small ship max. 2x
				r = r * ma / 3; // 1/3 of the original rate with same mass, min. 1/5 max. 2/3
//				player.consoleMessage("ILS slow steering with towed ship "+ato);//debug
			}
			
			var c = ph.cross(v).direction(); //set the plane where we should rotate in
			var q = p.orientation.rotate( c, -r );
			p.orientation = q;
		}
		
		this.$P = p.heading;

		if( npc ) {  //NPC ILS
                        var ang = 0, dist = 0, n = null;
                        var colldist = 2000; //check for collision within this distance
			var sp = p.maxSpeed;
                        var coll = false; //check for possible collision with another ship
                        var targets = this.$Targets; //filled by checkScanner() in timer
                        if( !targets && !(targets.length > 0) ) targets = [];
                        targets.push(player.ship);
                        for( var i = 0; i < targets.length; i++ ) {
                                n = targets[i];
                                if( n && n.isValid ) {
                                        dist = pp.distanceTo(n.position);
                                        if( dist < colldist ) { //is there a ship within 2km?
                                                ang = p.heading.angleTo(pp.subtract(n.position));
                                                //if( n == player.ship ) log(this.name, ang);//debug
                                                if( ang > 2 ) { //facing within 65 degree?
                                                        coll = n; //store the nearest dangerous ship
                                                        colldist = dist; //used below for desiredSpeed also
                                                }
                                        }
                                }
                        }
                        if( p.boundingBox.x > maxw && dd < noinjdist ) { //wide ship in the last 2km
				p.destination = ap; //AI will slow down near the ap
				if( dd > p.collisionRadius + 120 + dz/2 ) {
					p.desiredSpeed = Math.min(sp, 200); //approach slowly
				} else {
					ang = p.vectorUp.angleTo(dr); //angle between npc and dock
                                        var r = 0;
					if( this.$N || ang > 1.4 && ang < 1.74 ) { //check upside down also
						//right angle to start docking
						this.$N = true;//prevent stop after station rolled
						p.desiredSpeed = sp;
						p.destination = ap.add(ap.subtract(pp).multiply(1000));
                                                r = this.$S.roll * delta; //roll as the station
					} else {
						p.desiredSpeed = 0; //wait for the right angle
                                                //need roll especially at non rolling stations
                                                r = p.maxRoll/2 * delta;
                                                if( this.$S.roll > 0 ) r = -r; //roll opposite
					}
                                        p.orientation = p.orientation.rotate(ph, r);
                                        //player.consoleMessage(ang+" "+r);//debug
				}
			} else {
				if( dd > noinjdist ) { //use Injector far from the station
					if( p.injectorSpeedFactor > 0 )
						p.desiredSpeed = p.injectorSpeedFactor * sp;
					else p.desiredSpeed = 7 * sp; //for Oolite 1.80
				} else p.desiredSpeed = sp; //normal speed near the station
				//place destination farther to do not slow down when ap is near
				p.destination = ap.add(ap.subtract(pp).multiply(5));
				//this.$V.position = p.destination; //debug
			}
                        if( coll ) { //slow down when another ship is near, as slow as near
                                p.desiredSpeed = Math.min(p.desiredSpeed, 0.1*colldist);
                                //log(this.name, ang);//debug
                        }
			p.desiredRange = 0;
//			player.consoleMessage(p.desiredSpeed+" "+Math.round(p.speed));//debug
		} else { //player rotation before the dock if wide ship in the last 2km
                        if( ato < 0.005 && p.boundingBox.x > maxw && dd < noinjdist ) {
                                var ang = p.vectorUp.angleTo(dr); //angle between player ship and dock
                                var r = 0;
                                if( ang > 1.55 && ang < 1.6 ) { //check upside down also
                                        r = this.$S.roll * delta; //roll as the station
                                } else {
                                        r = p.maxRoll/2 * delta;
                                        if( this.$S.roll > 0 ) r = -r; //roll opposite
                                }
                                p.orientation = p.orientation.rotate(ph, r);
                                //player.consoleMessage(ang+" "+r);//debug
                        }
		}

	} else this.$ClearILS(); //no target or a new target is not station, etc.
}

this.$Timed = function() {
        var c = player.ship; //check around player
        if( this.$S ) c = this.$S; //check around last station if set
        if( !c || !c.isValid ) return;
        this.$Targets = c.checkScanner(true);//much faster than filteredEntities
}